//--------------------------------------------------------------------------
//
//   Copyright (c) 2002, Colin Granville
//
//   All rights reserved.
//
//   Redistribution and use in source and binary forms, with or
//   without modification, are permitted provided that the following 
//   conditions are met:
//
//      * Redistributions of source code must retain the above copyright 
//        notice, this list of conditions and the following disclaimer.
//
//      * Redistributions in binary form must reproduce the above 
//        copyright notice, this list of conditions and the following 
//        disclaimer in the documentation and/or other materials 
//        provided with the distribution.
//
//      * The name Colin Granville may not be used to endorse or promote 
//        products derived from this software without specific prior 
//        written permission.
//
//   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
//   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
//   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
//   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
//   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
//   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
//   OF THE POSSIBILITY OF SUCH DAMAGE.
//
//--------------------------------------------------------------------------

#include "string.h"
#include <string.h>
#include <iostream.h>
#include <ctype.h>

inline size_t min(size_t a,size_t b) {return (a>b ? b:a);}

string::InternalString string::null_string;

string string::add(const char *s1,int len1,const char *s2,int len2)
{
  string s;
  s.reserve(len1+len2);
  s.append(s1,len1);
  s.append(s2,len2);
  return s;
}

//****************************************************************

string operator+(char lhs, const string& rhs)
{
  string s;
  s.reserve(rhs.size()+1);
  s.append(1,lhs);
  s.append(rhs);
  return s;
}

//****************************************************************

string operator+(const string& lhs, char rhs)
{
  string s;
  s.reserve(lhs.size()+1);
  s.append(lhs);
  s.append(1,rhs);
  return s;
}

//****************************************************************

bool string::reserve_now(size_type n)
{
   if (p->hdr.count > 1 || n > capacity())
      {
      InternalString* old_p=p;
      p=allocate(n);
      if (p==&null_string)
         {
         p=old_p;
         return 0;
         }
      assign(old_p->data,old_p->hdr.size);         
      deallocate(old_p);
      }
   return 1;
}
//****************************************************************

bool string::midextend(size_type pos,int by)
{
   bool do_copy=(p->hdr.count > 1 || by+size() > capacity());
   InternalString* old_p=p;
   if (pos>size()) pos=size();
   if (do_copy)
      {
      p=allocate(old_p->hdr.size+by);
      if (p==&null_string)
         {
         p=old_p;
         return 0;
         }
      memmove(p->data,old_p->data,(by<0)?pos+by:pos);
      }
   memmove(p->data+pos+by,old_p->data+pos,old_p->hdr.size-pos);
   p->hdr.size=old_p->hdr.size+by;
   if (do_copy) deallocate(old_p);
   return 1;
}
//****************************************************************

string::InternalString* string::allocate(size_type n) 
{
  n=(((n+1)+7)&~7)+4; // allow room for trailing 0 
                         // first allocation 12 bytes then multiples of 8
                         // to allow for expansion
  InternalString* p=(InternalString *)new char[sizeof(InternalString::Hdr)+n];
  if (p) 
     {
     p->hdr.size=0;
     p->hdr.capacity=n-1; // capacity does not include room for trailing 0
                          // so that it is invisible to user
     p->hdr.count=1;
     }
  return (p)?p:&null_string;
}

//****************************************************************

void string::deallocate(InternalString* p)
{
   if (p != &null_string && --p->hdr.count==0 ) 
      { 
      delete [] p;
      }
}

//****************************************************************
//              constructors
//****************************************************************

string::string() : p(&null_string) {}

//****************************************************************

string::string(const string& str,size_type pos,size_type n)
 : p(&null_string)
{ assign(str,pos,n);}

//****************************************************************

string::string(const char* s) : p(&null_string) {assign(s,strlen(s));}

//****************************************************************

string::string(const char* s,size_type n) : p(&null_string) {assign(s,n);}

//****************************************************************

string::~string() {deallocate(p);}

//****************************************************************
//   capacity    
//****************************************************************

void string::reserve(size_type n)
{
  if (n <= capacity() && p->hdr.count == 1) return;
  InternalString* old_p=p;
  p=allocate(n);
  if (p==&null_string)
     {
     p=old_p;
     return;
     }
  assign(old_p->data,old_p->hdr.size);
  deallocate(old_p);  
}

//****************************************************************

void string::resize(size_type n, char c)
{
   if (reserve_now(n))
      {
      if (n > size()) memset(p->data+size(),c,n-size());
      p->hdr.size=n;
      }
}

//****************************************************************
//  modifiers
//****************************************************************

string& string::append(const string& str) {return append(str.p->data,str.size());}

//****************************************************************

string& string::append(const string& str, size_type pos,size_type n)
{
  if (pos>str.size()) pos=str.size();
  if (pos+n>str.size()) n=str.size()-pos;
  return append(str.p->data+pos,n);
}

//****************************************************************

string& string::append(const char* s, size_type n)
{
  if (n && reserve_now(size()+n))
     {
     memcpy(p->data+size(),s,n);
     n+=size();
     p->hdr.size=n;
     }
  return *this;
}

//****************************************************************

string& string::append(const char* s) {return append(s,strlen(s));} 

//****************************************************************

string& string::append(size_type n, char c) 
{
  resize(n+size(),c);
  return *this;
}

//****************************************************************

string& string::append(iterator first, iterator last)
{return append(first,last-first);}

//****************************************************************
//****************************************************************

string& string::assign(const string& str) 
{
  if (this!=&str) 
     {
     deallocate(p);
     p=str.p;
     p->hdr.count++;
     }
  return *this;
}

//****************************************************************

string& string::assign(const string& str,size_type pos,size_type n) 
{
  if (this==&str && pos==0 && n>=size()) return *this;
  if (pos > str.size()) pos=str.size();
  if (n > str.size()-pos) n=str.size()-pos;
  assign(str.data()+pos,n);
  return *this;
}

//****************************************************************

string& string::assign(const char* s) {return assign(s,strlen(s));}

//****************************************************************

string& string::assign(const char* s, size_type n)
{
  if (n!=npos)
     {
     InternalString* old_p=p;
     if (p->hdr.count > 1 || capacity() < n) p=allocate(n);
     if (p==&null_string)
        {
        p=old_p;
        return *this;
        }
     memmove(p->data,s,n);
     p->hdr.size=n;
     p->data[n]=0;
     if (p!=old_p) deallocate(old_p);
     }

  return *this;
}

//****************************************************************

string& string::assign(size_type n, char c)
{
  InternalString* old_p=p;

  if (p->hdr.count > 1 || capacity() < n) p=allocate(n);
  if (p==&null_string)
     {
     p=old_p;
     return *this;
     }
  memset(p->data,c,n);
  p->hdr.size=n;
  p->data[n]=0;
  if (p!=old_p) deallocate(old_p);

  return *this;
}

//****************************************************************
//****************************************************************

string& string::insert(size_type pos, const string& str)
{
  return insert(pos,str.p->data,str.size());
}

//****************************************************************

string& string::insert(size_type pos1, const string& str,size_type pos2, size_type n)
{
  if (pos2>str.length()) pos2=str.length();
  if (pos2+n>str.length()) n=str.length()-pos2;
  return insert(pos1,str.p->data+pos2,n);
}

//****************************************************************

string& string::insert(size_type pos, const char* s, size_type n)
{
  if (midextend(pos,n)) memcpy(p->data+pos,s,n);
  return *this;
}
 
//****************************************************************

string& string::insert(size_type pos, const char* s)
{
  return insert(pos,s,strlen(s));
}

//****************************************************************

string& string::insert(size_type pos, size_type n, char c)
{
  if (midextend(pos,n)) memset(p->data+pos,c,n);
  return *this;
}

//****************************************************************

string::iterator string::insert(iterator i, char c)
{
  size_type pos=i-p->data;
  if (midextend(pos,1)) *(p->data+pos)=c;
  return p->data+pos;
}

//****************************************************************

void string::insert(iterator i, size_type n, char c)
{
  size_type pos=i-p->data;
  if (midextend(pos,n)) memset(p->data+pos,c,n);
}

//****************************************************************

void string::insert(iterator i, iterator first, iterator last)
{
  insert(i-p->data,first,last-first);
}
//****************************************************************
//****************************************************************

string& string::erase(size_type pos, size_type n)
{
  if (pos<size())
     {
     int by=(n==npos || pos+n>size())?size()-pos:n;
     midextend(pos+by,-by);
     }
  return *this;
}

//****************************************************************

string::iterator string::erase(iterator position)
{
   size_type pos=position-p->data;
   erase(pos,1);
   return p->data+pos;
}

//****************************************************************

string::iterator string::erase(iterator first, iterator last)
{
   size_type pos=first-p->data;
   erase(pos,first-last);
   return p->data+pos;
}

//****************************************************************
//****************************************************************

string& string::replace(size_type pos1, size_type n1, const string& str)
{
  return replace(pos1,n1,str.p->data,str.size());
}

//****************************************************************

string& string::replace(size_type pos1, size_type n1, const string& str,
                        size_type pos2, size_type n2)
{
  if (pos2>str.size()) pos2=str.size();
  if (pos2+n2>str.size()) n2=str.size()-pos2;
  return replace(pos1,n1,str.p->data+pos2,n2);
}

//****************************************************************

string& string::replace(size_type pos, size_type n1, const char* s,size_type n2)
{
  if (pos>size()) pos=size();
  if (pos+n1>size()) n1=size()-pos;
  if (midextend(pos+n1,n2-n1)) memcpy(p->data+pos,s,n2);
  return *this;
}

//****************************************************************

string& string::replace(size_type pos, size_type n1, const char* s)
{
  return replace(pos,n1,s,strlen(s));
}

//****************************************************************

string& string::replace(size_type pos, size_type n1, size_type n2, char c)
{
  if (pos>size()) pos=size();
  if (pos+n1>size()) n1=size()-pos;
  if (midextend(pos+n1,n2-n1)) memset(p->data+pos,c,n2);
  return *this;
}

//****************************************************************

string& string::replace(iterator i1, iterator i2, const string& str)
{
  size_type pos=i1-p->data;
  return replace(pos,i2-i1,str.p->data,str.size());
}

//****************************************************************

string& string::replace(iterator i1, iterator i2, const char* s, size_type n)
{
  size_type pos=i1-p->data;
  return replace(pos,i2-i1,s,n);
}

//****************************************************************

string& string::replace(iterator i1, iterator i2, const char* s)
{
  size_type pos=i1-p->data;
  return replace(pos,i2-i1,s,strlen(s));
}

//****************************************************************

string& string::replace(iterator i1, iterator i2, size_type n, char c)
{
  size_type pos=i1-p->data;
  return replace(pos,i2-i1,n,c);
}

//****************************************************************

string& string::replace(iterator i1, iterator i2, iterator j1, iterator j2)
{
  size_type pos=i1-p->data;
  return replace(pos,i2-i1,j1,j2-j1);
}

//****************************************************************
//****************************************************************

string::size_type string::copy(char* s, size_type n, size_type pos) const
{
  if (pos>size()) pos=size();
  if (n>size()-pos) n=size()-pos;
  memcpy(s,p->data+pos,n);
  return n;
}

//****************************************************************
//****************************************************************

string::size_type string::find (const string& str, size_type pos) const
{
  return find(str.p->data,pos,str.size());
}

//****************************************************************

string::size_type string::find (const char* s, size_type pos, size_type n) const
{
  if (n && n < size() && pos<=size()-n)
     {
     char* from=p->data+pos;
     char* const to=from+size()-n;
     for (;from<=to;from++) 
         {
         if (*from==*s && memcmp(from,s,n)==0) return from-p->data;
         }
     }
  return npos;
}

//****************************************************************

string::size_type string::find (const char* s, size_type pos) const
{
  return find(s,pos,strlen(s));
}

//****************************************************************

string::size_type string::find (char c, size_type pos) const
{
  size_type len=size();
  for (;pos<len;pos++) {if (p->data[pos]==c) return pos;}
  return npos;
}

//****************************************************************
//****************************************************************

string::size_type string::rfind(const string& str, size_type pos) const
{
  return rfind(str.p->data,pos,str.size());
}

//****************************************************************

string::size_type string::rfind(const char* s, size_type pos, size_type n) const
{
  if (pos>size()) pos=size();
  if (n && n <= pos)
     {
     char* from=p->data+pos-n;
     char* const to=p->data;
     for(;from>=to;from--)
         {
         if (*from==*s && memcmp(from,s,n)==0) return from-p->data;
         }
     }
  return npos;
}

//****************************************************************

string::size_type string::rfind(const char* s, size_type pos) const
{
  return rfind(s,pos,strlen(s));
}

//****************************************************************

string::size_type string::rfind(char c, size_type pos) const
{
  if (pos>size()) pos=size();
  if (pos)
     {
     char* from=p->data+pos-1;
     char* const to=p->data;
     for(;from>=to;from--)
         {
         if (*from==c) return from-p->data;
         }
     }
  return npos;
}

//****************************************************************
//****************************************************************

string::size_type string::find_first_of(const string& str, size_type pos) const
{
  return find_first_of(str.p->data,pos,str.size());
}

//****************************************************************

string::size_type string::find_first_of(const char* s, size_type pos, size_type n) const
{
  if (pos<size())
     {
     char* from=p->data+pos;
     char* const to=from+size();
     for (;from<to;from++) 
         {
         if (memchr(s,*from,n)!=NULL) return from-p->data;
         }
     }
  return npos;
}

//****************************************************************

string::size_type string::find_first_of(const char* s, size_type pos) const
{
  return find_first_of(s,pos,strlen(s));
}

//****************************************************************

string::size_type string::find_first_of(char c, size_type pos) const
{
  return find(c,pos);
}

//****************************************************************
//****************************************************************

string::size_type string::find_last_of (const string& str,size_type pos) const
{
  return find_last_of(str.p->data,pos,str.size());
}

//****************************************************************

string::size_type string::find_last_of (const char* s, size_type pos, size_type n) const
{
  if (pos>size()) pos=size();
  if (pos)
     {
     char* from=p->data+pos-1;
     char* const to=p->data;
     for(;from>=to;from--)
         {
         if (memchr(s,*from,n)!=NULL) return from-p->data;
         }
     }
  return npos;
}

//****************************************************************

string::size_type string::find_last_of (const char* s, size_type pos) const
{
  return find_last_of(s,pos,strlen(s));
}

//****************************************************************

string::size_type string::find_last_of (char c, size_type pos) const
{
   return rfind(c,pos);
}

//****************************************************************
//****************************************************************

string::size_type string::find_first_not_of(const string& str, size_type pos) const
{
  return find_first_not_of(str.p->data,pos,str.size());
}

//****************************************************************

string::size_type string::find_first_not_of(const char* s, size_type pos, size_type n) const
{
  if (pos<size())
     {
     char* from=p->data+pos;
     char* const to=from+size();
     for (;from<to;from++) 
         {
         if (memchr(s,*from,n)==NULL) return from-p->data;
         }
     }
  return npos;
}

//****************************************************************

string::size_type string::find_first_not_of(const char* s, size_type pos) const
{
   return find_first_not_of(s,pos,strlen(s));
}

//****************************************************************

string::size_type string::find_first_not_of(char c, size_type pos) const
{
   size_type len=size();
   for (;pos<len;pos++) {if (p->data[pos]!=c) return pos;}
   return npos;
}
//****************************************************************
//****************************************************************

string::size_type string::find_last_not_of (const string& str, size_type pos) const
{
  return find_last_not_of(str.p->data,pos,str.size());
}


//****************************************************************

string::size_type string::find_last_not_of (const char* s, size_type pos, size_type n) const
{
  if (pos>size()) pos=size();
  if (n && n <= pos)
     {
     char* from=p->data+pos-n;
     char* const to=p->data;
     for(;from>=to;from--)
         {
         if (memchr(s,*from,n)==NULL) return from-p->data;
         }
     }
  return npos;
}

//****************************************************************

string::size_type string::find_last_not_of (const char* s, size_type pos) const
{
  return find_last_not_of(s,pos,strlen(s));
}

//****************************************************************

string::size_type string::find_last_not_of (char c, size_type pos) const
{
  if (pos>size()) pos=size();
  if (pos)
     {
     char* from=p->data+pos-1;
     char* const to=p->data;
     for(;from>=to;from--)
         {
         if (*from!=c) return from-p->data;
         }
     }
  return npos;
}

//****************************************************************
//****************************************************************

int string::compare(const string& str) const
{
  int len=(length()<str.length())?length():str.length();
  int comp= memcmp(data(),str.data(),len);
  return (comp==0)?length()-str.length():comp;
}

//****************************************************************

int string::compare(size_type pos1, size_type n1, const string& str) const
{
  return compare(pos1,n1,str,0,str.length());
}

//****************************************************************
int string::compare(size_type pos1, size_type n1, const string& str,
                                        size_type pos2, size_type n2) const
{
  if (pos1>length()) pos1=length();
  if (pos1+n1>length()) n1=length()-pos1;
  if (pos2>str.length()) pos2=str.length();
  if (pos2+n2>str.length()) n2=str.length()-pos2;
  int comp= memcmp(data()+pos1,str.data()+pos2,min(n1,n2));
  return (comp==0)?n1-n2:comp;
}
  
//****************************************************************

int string::compare(const char* s) const
{
  char* from=p->data;
  char* to=p->data+size();
  char a,b;
  while (from<to)
        {
        a=*from++;
        b=*s++;
        if (a!=b || b==0) return (a==b)?1:a-b;
        }
  return -*s;
}

//****************************************************************

int string::compare(size_type pos1, size_type n1, const char* s, size_type n2) const
{
  if (pos1>length()) pos1=length();
  if (pos1+n1>length()) n1=length()-pos1;
  int comp=memcmp(data()+pos1,s,(n1>n2)?n2:n1);
  return (comp==0)?n1-n2:comp;
}

//****************************************************************
//****************************************************************

ostream& operator<<(ostream& out,const string& str)
{
  out.write(str.data(),str.size());
  out.osfx();
  return out;
}

//****************************************************************

istream& operator>>(istream& in,string& str)
{
  if (in.ipfx(0))
     {
     string::size_type len=(in.width())?in.width():512;
     char p[512];
     string::size_type pos=0;
     int c;
     while (pos<len)
           {
           c=in.peek();
           if (c==EOF || isspace(c)) break;
           in.get();
           p[pos++]=c;
           }
     str.erase();
     if (pos) str.assign(p,pos);
     in.width(0);
     }
  return in;
}


//****************************************************************
istream& getline(istream& in, string& str, char delim)
{
  if (in.ipfx(1))
     {
     char p[512];
     string::size_type pos=0;
     int c;
     while (pos<512)
           {
           c=in.get();
           if (c==EOF || c==delim) break;
           p[pos++]=c;
           }
     str.erase();
     if (pos) str.assign(p,pos);
     }
  return in;
}

