//--------------------------------------------------------------------------
//
//   Copyright (c) 2002, Colin Granville
//
//   All rights reserved.
//
//   Redistribution and use in source and binary forms, with or
//   without modification, are permitted provided that the following
//   conditions are met:
//
//      * Redistributions of source code must retain the above copyright
//        notice, this list of conditions and the following disclaimer.
//
//      * Redistributions in binary form must reproduce the above
//        copyright notice, this list of conditions and the following
//        disclaimer in the documentation and/or other materials
//        provided with the distribution.
//
//      * The name Colin Granville may not be used to endorse or promote
//        products derived from this software without specific prior
//        written permission.
//
//   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
//   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
//   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
//   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
//   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
//   OF THE POSSIBILITY OF SUCH DAMAGE.
//
//--------------------------------------------------------------------------

#include "Object.h"
#include "GString.h"
#include "GuiWindow.h"
#include "GuiTargets.h"
#include "GuiTask.h"
#include "Document.h"
#include "DocView.h"
#include "flex.h"
#include <fstream>
#include <strstream>
#include <stdlib.h>
#include <stdarg.h>
#include "gtypes.h"
#include "Node.h"
#include "Error.h"
#include "GlobalParams.h"
#include "UserEvents.h"
#include "GuiIconbar.h"
#include "GuiFlexGlobal.h"
#include "DynamicCast.h"
#include "unixlib/local.h"

int __riscosify_control = __RISCOSIFY_NO_PROCESS;

Document* makeDocument(Node& application,const char* file_name);
bool       makeDocView(Document*,const DocViewChoices&,int,int);

// first part is the xpdf version number
#ifndef __APCS_32
#define  VERSION "3.02.1.24"
#else
#define  VERSION "3.02.1.24"
#endif

class Application : public Node
{
  public:
    Application(int argc,char** argv);
    virtual Node* removeChild(Node*);
    virtual bool  hasChildNodes()                {return 1;}

  private:
    GuiIconbar iconbar;

    virtual ~Application() {}

    bool createDocument(const char* filename);

    GUI_DECLARE_EVENT_TARGETS(Application);

    GuiToolboxTarget userQuitTarget;
    Claim userQuit(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget showHelpTarget;
    Claim showHelp(GuiToolboxEvent&,const GuiIdBlock&);


    GuiMessageTarget messageQuitTarget;
    Claim messageQuit(GuiWimpMessage&);

    GuiMessageTarget dataOpenTarget;
    Claim dataOpen(GuiWimpMessage&);

    GuiMessageTarget dataLoadTarget;
    Claim dataLoad(GuiWimpMessage&);
};

//*************************************************************************

Application::Application(int argc,char** argv)
  : iconbar("Iconbar"),
    userQuitTarget(0,User_Quit,this,(Claim (Application::*)(GuiToolboxEvent&, const GuiIdBlock&))&Application::userQuit),
    showHelpTarget(0,User_ShowHelp,this,(Claim (Application::*)(GuiToolboxEvent&, const GuiIdBlock&))&Application::showHelp),
    messageQuitTarget(GuiWimp_MQuit,this,(Claim (Application::*)(GuiWimpMessage&))&Application::messageQuit),
    dataOpenTarget(GuiWimp_MDataOpen,this,(Claim (Application::*)(GuiWimpMessage&))&Application::dataOpen),
    dataLoadTarget(GuiWimp_MDataLoad,this,(Claim (Application::*)(GuiWimpMessage&))&Application::dataLoad)
{
  char* file_name=0;
  if (argc>1) file_name=argv[1];

  if (file_name==0)  iconbar.show();

  if (!createDocument(file_name) && !iconbar.isShowing()) exit(EXIT_SUCCESS);
}

//*************************************************************************

bool Application::createDocument(const char* filename)
{
   DocViewChoices choices;
   return makeDocView(makeDocument(*this,filename),choices);
}

//*************************************************************************

Node* Application::removeChild(Node* n)
{
  Node::removeChild(n);
  if (getFirstChild()==0 && !iconbar.isShowing())
  {
    delete this;
    exit(EXIT_SUCCESS);
  }
  return n;
}

//*************************************************************************

Claim Application::userQuit(GuiToolboxEvent&,const GuiIdBlock&)
{
  delete this;
  exit(EXIT_SUCCESS);
  return CLAIM;
}

//*************************************************************************

Claim Application::showHelp(GuiToolboxEvent&,const GuiIdBlock&)
{
  _swix(OS_CLI,_IN(0),"filer_run <pdf$dir>.!help");
  return CLAIM;
}

//*************************************************************************

Claim Application::messageQuit(GuiWimpMessage&)
{
  delete this;
  exit(EXIT_SUCCESS);
  return CLAIM;
}

//*************************************************************************

Claim Application::dataOpen(GuiWimpMessage& mess)
{
//  bool claim=0;
  switch (mess.data.dataOpen.fileType)
  {
    case 0xadf: mess.acknowledge();
                createDocument(mess.data.dataOpen.pathName);
                return CLAIM;
  }
  return DONT_CLAIM;
}

//*************************************************************************

Claim Application::dataLoad(GuiWimpMessage& mess)
{
  if (mess.data.dataLoad.destination.windowHandle!=-2) return DONT_CLAIM;

  if (createDocument(mess.data.dataLoad.leafName))
  {
      mess.reply(GuiWimp_MDataLoadAck);
  }
  return CLAIM;
}

//*************************************************************************
//*************************************************************************
//*************************************************************************
class ProgramInfo
{
  public:
    ProgramInfo();
    ~ProgramInfo();
  private:
    GuiWindow       info;
    GuiDisplayField version;

    GUI_DECLARE_EVENT_TARGETS(ProgramInfo);

    GuiToolboxTarget atbs_target;
    Claim atbs(GuiToolboxEvent&,const GuiIdBlock&);
};

ProgramInfo::ProgramInfo()
  : info("ProgInfo"),
    version(info,9),
    atbs_target(&info,0x82880,this,(Claim (ProgramInfo::*)(GuiToolboxEvent&, const GuiIdBlock&))&ProgramInfo::atbs)
{
}
//*****************************************************************

ProgramInfo::~ProgramInfo()
{
}

//*****************************************************************

Claim ProgramInfo::atbs(GuiToolboxEvent&,const GuiIdBlock&)
{
  char buf[64];
  ostrstream out(buf,sizeof(buf));
  out << VERSION << " (";
  if (__DATE__[4] != ' ') out <<__DATE__[4];
  out << __DATE__[5] << ' ';
  out.write(__DATE__,3);
  out <<' ' << (__DATE__+7) << ')' << ends;
  version.setValue(out.str());
  return CLAIM;
}

//*************************************************************************
//*************************************************************************
//*************************************************************************
int main(int argc,char** argv)
{
  globalParams = new GlobalParams("");
  int debug= (argc>2 && argv[2][0]=='1');
  globalParams->setErrQuiet(debug?0:1);
  globalParams->setPrintCommands(debug?1:0);
  globalParams->setupBaseFonts(NULL);
  guiTaskInitialise("<PDF$Dir>");
  new Application(argc,argv);
  ProgramInfo info;
  for (;;)  GuiRegister::poll();
}
